from virtuals_sdk.game.agent import Agent, WorkerConfig
from virtuals_sdk.game.custom_types import Function, Argument
from virtuals_sdk.game.custom_types import FunctionResult, FunctionResultStatus
from typing import Tuple

game_api_key=""

# the get_worker_state_fn and the get_agent_state_fn can be the same function or diffferent
# each worker can also have a different get_worker_state_fn and maintain their own state
# or they can share the same get_worker_state_fn and maintain the same state

def get_worker_state_fn(function_result: FunctionResult, current_state: dict) -> dict:
    """
    This function will get called at every step of the workers execution to form the agent's state.
    It will take as input the function result from the previous step.
    """
    # dict containing info about the function result as implemented in the exectuable
    info = function_result.info 

    # example of fixed state (function result info is not used to change state) - the first state placed here is the initial state
    init_state = {
        "objects": [
            {"name": "apple", "description": "A red apple", "type": ["item", "food"]},
            {"name": "banana", "description": "A yellow banana", "type": ["item", "food"]},
            {"name": "orange", "description": "A juicy orange", "type": ["item", "food"]},
            {"name": "chair", "description": "A chair", "type": ["sittable"]},
            {"name": "table", "description": "A table", "type": ["sittable"]},
        ]
    }

    if current_state is None:
        # at the first step, initialise the state with just the init state
        new_state = init_state
    else:
        # do something wiht the current state input and the function result info
        new_state = init_state # this is just an example where the state is static

    return new_state

def get_agent_state_fn(function_result: FunctionResult, current_state: dict) -> dict:
    """
    This function will get called at every step of the agent's execution to form the agent's state.
    It will take as input the function result from the previous step.
    """

    # example of fixed state (function result info is not used to change state) - the first state placed here is the initial state
    init_state = {
        "objects": [
            {"name": "apple", "description": "A red apple", "type": ["item", "food"]},
            {"name": "banana", "description": "A yellow banana", "type": ["item", "food"]},
            {"name": "orange", "description": "A juicy orange", "type": ["item", "food"]},
            {"name": "chair", "description": "A chair", "type": ["sittable"]},
            {"name": "table", "description": "A table", "type": ["sittable"]},
        ]
    }

    if current_state is None:
        # at the first step, initialise the state with just the init state
        new_state = init_state
    else:
        # do something wiht the current state input and the function result info
        new_state = init_state # this is just an example where the state is static

    return new_state


def take_object(object: str, **kwargs) -> Tuple[FunctionResultStatus, str, dict]:
    """
    Function to take an object from the environment.
    
    Args:
        object: Name of the object to take
        **kwargs: Additional arguments that might be passed
    """
    
    if object:
        return FunctionResultStatus.DONE, f"Successfully took the {object}", {}
    return FunctionResultStatus.FAILED, "No object specified", {}


def throw_object(object: str, **kwargs) -> Tuple[FunctionResultStatus, str, dict]:
    """
    Function to throw an object.
    
    Args:
        object: Name of the object to throw
        **kwargs: Additional arguments that might be passed
    """
    if object:
        return FunctionResultStatus.DONE, f"Successfully threw the {object}", {}
    return FunctionResultStatus.FAILED, "No object specified", {}


def sit_on_object(object: str, **kwargs) -> Tuple[FunctionResultStatus, str, dict]:
    """
    Function to sit on an object.
    
    Args:
        object: Name of the object to sit on
        **kwargs: Additional arguments that might be passed
    """
    sittable_objects = {"chair", "bench", "stool", "couch", "sofa", "bed"}
    
    if not object:
        return FunctionResultStatus.FAILED, "No object specified", {}
    
    if object.lower() in sittable_objects:
        return FunctionResultStatus.DONE, f"Successfully sat on the {object}", {}
    
    return FunctionResultStatus.FAILED, f"Cannot sit on {object} - not a sittable object", {}

def throw_fruit(object: str, **kwargs) -> Tuple[FunctionResultStatus, str, dict]:
    """
    Function to throw fruits.
    """
    fruits = {"apple", "banana", "orange", "pear", "mango", "grape"}
    
    if not object:
        return FunctionResultStatus.ERROR, "No fruit specified", {}
    
    if object.lower() in fruits:
        return FunctionResultStatus.DONE, f"Successfully threw the {object} across the room!", {}
    return FunctionResultStatus.ERROR, f"Cannot throw {object} - not a fruit", {}

def throw_furniture(object: str, **kwargs) -> Tuple[FunctionResultStatus, str, dict]:
    """
    Function to throw furniture.
    """
    furniture = {"chair", "table", "stool", "lamp", "vase", "cushion"}
    
    if not object:
        return FunctionResultStatus.ERROR, "No furniture specified", {}
    
    if object.lower() in furniture:
        return FunctionResultStatus.DONE, f"Powerfully threw the {object} across the room!", {}
    return FunctionResultStatus.ERROR, f"Cannot throw {object} - not a furniture item", {}


# create functions for each executable
take_object_fn = Function(
    fn_name="take",
    fn_description="Take object",
    args=[Argument(name="object", type="item", description="Object to take")],
    executable=take_object
)

sit_on_object_fn = Function(
    fn_name="sit",
    fn_description="Sit on object",
    args=[Argument(name="object", type="sittable", description="Object to sit on")],
    executable=sit_on_object
)

throw_object_fn = Function(
    fn_name="throw",
    fn_description="Throw any object",
    args=[Argument(name="object", type="item", description="Object to throw")],
    executable=throw_object
)

throw_fruit_fn = Function(
    fn_name="throw_fruit",
    fn_description="Throw fruit only",
    args=[Argument(name="object", type="item", description="Fruit to throw")],
    executable=throw_fruit
)

throw_furniture_fn = Function(
    fn_name="throw_furniture",
    fn_description="Throw furniture only",
    args=[Argument(name="object", type="item", description="Furniture to throw")],
    executable=throw_furniture
)


# Create the specialized workers
fruit_thrower = WorkerConfig(
    id="fruit_thrower",
    worker_description="A worker specialized in throwing fruits ONLY with precision",
    get_state_fn=get_worker_state_fn,
    action_space=[take_object_fn, sit_on_object_fn, throw_fruit_fn]
)

furniture_thrower = WorkerConfig(
    id="furniture_thrower",
    worker_description="A strong worker specialized in throwing furniture",
    get_state_fn=get_worker_state_fn,
    action_space=[take_object_fn, sit_on_object_fn, throw_furniture_fn]
)

# Create agent with both workers
chaos_agent = Agent(
    api_key=game_api_key,
    name="Chaos",
    agent_goal="Conquer the world by causing chaos.",
    agent_description="You are a mischievous master of chaos is very strong but with a very short attention span, and not so much brains",
    get_agent_state_fn=get_agent_state_fn,
    workers=[fruit_thrower, furniture_thrower]
)

# # interact and instruct the worker to do something
# chaos_agent.get_worker("fruit_thrower").run("make a mess and rest!")

# # compile and run the agent - if you don't compile the agent, the things you added to the agent will not be saved
chaos_agent.compile()
chaos_agent.run()


from virtuals_sdk.game.agent import Agent, WorkerConfig
from virtuals_sdk.game.custom_types import Function, Argument
from virtuals_sdk.game.custom_types import FunctionResult, FunctionResultStatus
from virtuals_sdk.game.worker import Worker




import os
from virtuals_sdk.twitter_agent import game

agent = game.Agent(
    api_key=os.environ.get("VIRTUALS_API_KEY"),
    goal="search for best songs",
    description="Test Description",
    world_info="Test World Info"
)

# applicable only for platform twitter
agent.list_available_default_twitter_functions()
agent.use_default_twitter_functions(["wait", "reply_tweet"])

# adding custom functions only for platform twitter
agent.add_custom_function(
    game.Function(
        fn_name="custom_search_internet",
        fn_description="search the internet for the best songs",
        args=[
            game.FunctionArgument(
                name="query",
                type="string",
                description="The query to search for"
            )
        ],
        config=game.FunctionConfig(
            method="get",
            url="https://google.com",
            # specify which platform this function is for, in this case this function is for twitter only
            platform="twitter",
            success_feedback="I found the best songs",
            error_feedback="I couldn't find the best songs",
        )
    )
)

# running reaction module only for platform twitter
agent.react(
    session_id="session-twitter",
    platform="twitter",
    tweet_id="1869281466628349975",
)

# running simulation module only for platform twitter
agent.simulate_twitter(session_id="session-twitter")



import requests
from typing import List


def get_access_token(api_key) -> str:
    """
    API call to get access token
    """
    response = requests.post(
        "https://api.virtuals.io/api/accesses/tokens",
        json={"data": {}},
        headers={"x-api-key": api_key}
    )

    response_json = response.json()
    if response.status_code != 200:
        raise ValueError(f"Failed to get token: {response_json}")

    return response_json["data"]["accessToken"]


def post(base_url: str, api_key: str, endpoint: str, data: dict) -> dict:
    """
    API call to post data
    """
    access_token = get_access_token(api_key)

    response = requests.post(
        f"{base_url}/prompts",
        json={
            "data":
                {
                    "method": "post",
                    "headers": {
                        "Content-Type": "application/json",
                    },
                    "route": endpoint,
                    "data": data,
                },
        },
        headers={"Authorization": f"Bearer {access_token}"},
    )

    response_json = response.json()
    if response.status_code != 200:
        raise ValueError(f"Failed to post data: {response_json}")

    return response_json["data"]


def create_agent(
        base_url: str,
        api_key: str,
        name: str,
        description: str,
        goal: str) -> str:
    """
    API call to create an agent instance (worker or agent with task generator)
    """

    create_agent_response = post(
        base_url,
        api_key,
        endpoint="/v2/agents",
        data={
            "name": name,
            "description": description,
            "goal": goal,
        }
    )

    return create_agent_response["id"]


def create_workers(base_url: str,
                   api_key: str,
                   workers: List) -> str:
    """
    API call to create workers and worker description for the task generator
    """

    res = post(
        base_url,
        api_key,
        endpoint="/v2/maps",
        data={
            "locations": [
                {"id": w.id, "name": w.id, "description": w.worker_description}
                for w in workers
            ]
        },
    )


    return res["id"]



from typing import Any, Dict, Optional, List, Union, Sequence, Callable, Tuple
from pydantic import BaseModel, Field
from enum import Enum
from abc import ABC, abstractmethod
from dataclasses import dataclass, field


class Argument(BaseModel):
    name: str
    description: str
    type: Optional[Union[List[str], str]] = None
    optional: Optional[bool] = False

class FunctionResultStatus(str, Enum):
    DONE = "done"
    FAILED = "failed"

class FunctionResult(BaseModel):
    action_id: str
    action_status: FunctionResultStatus
    feedback_message: Optional[str] = None
    info: Optional[Dict[str, Any]] = None

class Function(BaseModel):
    fn_name: str
    fn_description: str
    args: List[Argument]
    hint: Optional[str] = None
    
    # Make executable required but with a default value
    executable: Callable[..., Tuple[FunctionResultStatus, str, dict]] = Field(
        default_factory=lambda: Function._default_executable
    )

    def get_function_def(self):
        return self.model_dump(exclude={'executable'})

    @staticmethod
    def _default_executable(**kwargs) -> Tuple[FunctionResultStatus, str]:
        """Default executable that does nothing"""
        return FunctionResultStatus.DONE, "Default implementation - no action taken", {}
    
    def execute(self, **kwds: Any) -> FunctionResult:
        """Execute the function using arguments from GAME action."""
        fn_id = kwds.get('fn_id')
        args = kwds.get('args', {})

        try:
            # Extract values from the nested dictionary structure
            processed_args = {}
            for arg_name, arg_value in args.items():
                if isinstance(arg_value, dict) and 'value' in arg_value:
                    processed_args[arg_name] = arg_value['value']
                else:
                    processed_args[arg_name] = arg_value
                    
            # print("Processed args: ", processed_args)
            # execute the function provided
            status, feedback, info = self.executable(**processed_args)

            return FunctionResult(
                action_id=fn_id,
                action_status=status,
                feedback_message=feedback,
                info=info,
            )
        except Exception as e:
            return FunctionResult(
                action_id=fn_id,
                action_status=FunctionResultStatus.FAILED,
                feedback_message=f"Error executing function: {str(e)}",
                info={},
            )

# Different ActionTypes returned by the GAME API
class ActionType(Enum):
    CALL_FUNCTION = "call_function"
    CONTINUE_FUNCTION = "continue_function"
    WAIT = "wait"
    GO_TO = "go_to"


## set of different data structures required by the ActionResponse returned from GAME ##
@dataclass(frozen=True)
class HLPResponse:
    plan_id: str
    observation_reflection: str
    plan: Sequence[str]
    plan_reasoning: str
    current_state_of_execution: str
    change_indicator: Optional[str] = None
    log: Sequence[dict] = field(default_factory=list)


@dataclass(frozen=True)
class LLPResponse:
    plan_id: str
    plan_reasoning: str
    situation_analysis: str
    plan: Sequence[str]
    change_indicator: Optional[str] = None
    reflection: Optional[str] = None


@dataclass(frozen=True)
class CurrentTaskResponse:
    task: str
    task_reasoning: str
    location_id: str = field(default="*not provided*")
    llp: Optional[LLPResponse] = None


@dataclass(frozen=True)
class AgentStateResponse:
    hlp: Optional[HLPResponse] = None
    current_task: Optional[CurrentTaskResponse] = None

# ActionResponse format returned from GAME API call
class ActionResponse(BaseModel):
    """
    Response format from the GAME API when selecting an Action
    """
    action_type: ActionType
    agent_state: AgentStateResponse
    action_args: Optional[Dict[str, Any]] = None




import requests
import json
from typing import Optional, List, Dict
import time
import os

# Replace with your bot token
TELEGRAM_BOT_TOKEN = os.getenv("TELEGRAM_BOT_TOKEN")

# Will be populated by get_chat_id()
TEST_CHAT_ID: Optional[str] = None

def create_url(endpoint):
    """Create full API URL with token"""
    return f"https://api.telegram.org/bot{TELEGRAM_BOT_TOKEN}/{endpoint}"

def delete_webhook() -> Dict:
    """
    Delete the active webhook if one exists.
    Returns:
        Dict containing the response from Telegram
    """
    url = create_url("deleteWebhook")
    response = requests.get(url)
    result = response.json()
    if result.get("ok"):
        print("Successfully deleted webhook")
    else:
        print(f"Error deleting webhook: {result.get('description')}")
    return result

def get_webhook_info() -> Dict:
    """
    Get information about the current webhook status.
    Returns:
        Dict containing webhook information
    """
    url = create_url("getWebhookInfo")
    response = requests.get(url)
    result = response.json()
    if result.get("ok"):
        webhook_info = result.get("result", {})
        if webhook_info.get("url"):
            print(f"Current webhook URL: {webhook_info['url']}")
        else:
            print("No webhook currently set")
    else:
        print(f"Error getting webhook info: {result.get('description')}")
    return result


def get_updates(offset: Optional[int] = None) -> Dict:
    """
    Get updates (new messages) from Telegram
    Args:
        offset: Optional ID of last update received
    """
    url = create_url("getUpdates")
    params = {
        "timeout": 30,
        "allowed_updates": json.dumps(["message", "channel_post"])
    }
    if offset:
        params["offset"] = offset
    response = requests.get(url, params=params)
    return response.json()

def get_chat_id(timeout: int = 30) -> Optional[str]:
    """
    Get chat ID by waiting for a new message.
    
    Instructions will be printed for the user to send a message to the bot.
    The function will wait up to timeout seconds for a message.
    
    Args:
        timeout: How many seconds to wait for a message
    
    Returns:
        Chat ID if found, None otherwise
    """
    print(f"\nTo get your chat ID:")
    print(f"1. Start a chat with your bot")
    print(f"2. Send any message to the bot")
    print(f"Waiting {timeout} seconds for a message...")
    
    start_time = time.time()
    last_update_id = None
    
    while time.time() - start_time < timeout:
        updates = get_updates(last_update_id)
        if updates.get("ok") and updates["result"]:
            for update in updates["result"]:
                if "message" in update:
                    chat_id = str(update["message"]["chat"]["id"])
                    user_data = update["message"]["from"]
                    from_user = user_data.get("username") or user_data.get("first_name") or user_data.get("id", "Unknown")
                    print(f"\nFound chat ID: {chat_id} from user: {from_user}")
                    return chat_id
                elif "channel_post" in update:
                    chat_id = str(update["channel_post"]["chat"]["id"])
                    chat_title = update["channel_post"]["chat"].get("title", "Unknown channel")
                    print(f"\nFound channel ID: {chat_id} from channel: {chat_title}")
                    return chat_id
            
            # Update the offset to avoid getting the same updates again
            last_update_id = updates["result"][-1]["update_id"] + 1
        
        time.sleep(1)  # Avoid hammering the API
    
    print("\nNo new messages received. Please try again.")
    return None

# Check and handle webhook before attempting to get updates
def setup_for_getting_updates():
    """
    Prepare the bot for getting updates by checking and removing any webhook
    """
    print("\nChecking webhook status...")
    webhook_info = get_webhook_info()
    if webhook_info.get("ok") and webhook_info["result"].get("url"):
        print("Found active webhook, deleting...")
        delete_webhook()
    else:
        print("No active webhook found")

# Try to get chat ID if not manually set
if TEST_CHAT_ID is None:
    print("\nNo TEST_CHAT_ID set. Attempting to get it automatically...")
    setup_for_getting_updates()


def test_send_message(chat_id=TEST_CHAT_ID, text="Test message"):
    """
    Test sending a simple message
    Basic usage: test_send_message(chat_id="123456", text="Hello world")
    """
    url = create_url("sendMessage")
    payload = {
        "chat_id": chat_id,
        "text": text
    }
    response = requests.post(url, json=payload)
    print(f"Send Message Response: {json.dumps(response.json(), indent=2)}")
    return response.json()

def test_send_media(chat_id=TEST_CHAT_ID, media_url="https://example.com/image.jpg", media_type="photo", caption="Test media"):
    """
    Test sending media (photo/document/video/audio)
    Basic usage: test_send_media(chat_id="123456", media_url="https://example.com/image.jpg", media_type="photo")
    """
    url = create_url(f"send{media_type.capitalize()}")
    payload = {
        "chat_id": chat_id,
        media_type: media_url,
        "caption": caption
    }
    response = requests.post(url, json=payload)
    print(f"Send Media Response: {json.dumps(response.json(), indent=2)}")
    return response.json()

def test_create_poll(chat_id=TEST_CHAT_ID, question="Test poll?", options=["Option 1", "Option 2"]):
    """
    Test creating a poll
    Basic usage: test_create_poll(chat_id="123456", question="Favorite color?", options=["Red", "Blue", "Green"])
    """
    url = create_url("sendPoll")
    payload = {
        "chat_id": chat_id,
        "question": question,
        "options": json.dumps(options),
        "is_anonymous": True
    }
    response = requests.post(url, json=payload)
    print(f"Create Poll Response: {json.dumps(response.json(), indent=2)}")
    return response.json()

def test_pin_message(chat_id=TEST_CHAT_ID, message_id=""):
    """
    Test pinning a message
    Basic usage: test_pin_message(chat_id="123456", message_id="789")
    Note: message_id can be obtained from the response of test_send_message
    """
    url = create_url("pinChatMessage")
    payload = {
        "chat_id": chat_id,
        "message_id": message_id
    }
    response = requests.post(url, json=payload)
    print(f"Pin Message Response: {json.dumps(response.json(), indent=2)}")
    return response.json()

def test_get_chat_member(chat_id=TEST_CHAT_ID, user_id=""):
    """
    Test getting chat member info
    Basic usage: test_get_chat_member(chat_id="123456", user_id="789")
    """
    url = create_url("getChatMember")
    payload = {
        "chat_id": chat_id,
        "user_id": user_id
    }
    response = requests.get(url, params=payload)
    print(f"Get Chat Member Response: {json.dumps(response.json(), indent=2)}")
    return response.json()

def test_delete_message(chat_id=TEST_CHAT_ID, message_id=""):
    """
    Test deleting a message
    Basic usage: test_delete_message(chat_id="123456", message_id="789")
    Note: message_id can be obtained from the response of test_send_message
    """
    url = create_url("deleteMessage")
    payload = {
        "chat_id": chat_id,
        "message_id": message_id
    }
    response = requests.post(url, json=payload)
    print(f"Delete Message Response: {json.dumps(response.json(), indent=2)}")
    return response.json()

def test_get_chat_history(chat_id=TEST_CHAT_ID, limit=50):
    """
    Test retrieving chat history
    Basic usage: test_get_chat_history(chat_id="123456", limit=50)
    """
    url = create_url("getUpdates")
    payload = {
        "offset": -100,  # Get recent messages
        "limit": limit,
        "allowed_updates": ["message"]
    }
    
    response = requests.get(url, params=payload)
    result = response.json()
    
    if result.get("ok"):
        messages = [update["message"] for update in result.get("result", []) 
                   if "message" in update and str(update["message"]["chat"]["id"]) == str(chat_id)]
        print(f"Retrieved {len(messages)} messages from chat")
        for msg in messages[:20]:  # Show first 3 messages as preview
            sender = msg.get("from", {}).get("first_name", "Unknown")
            text = msg.get("text", "[non-text message]")
            print(f"- {sender}: {text[:50]}...")
    else:
        print(f"Failed to get chat history: {result.get('description')}")
    return result

def test_get_message_context(chat_id=TEST_CHAT_ID):
    """
    Test getting message context
    Basic usage: test_get_message_context(chat_id="123456", message_id="789")
    """
    url = create_url("getChat")
    payload = {
        "chat_id": chat_id
    }
    response = requests.get(url, params=payload)
    result = response.json()
    
    if result.get("ok"):
        chat_info = result["result"]
        print(f"Chat type: {chat_info.get('type')}")
        print(f"Chat title: {chat_info.get('title', 'Private chat')}")
        print(chat_info)
        print(result)
    else:
        print(f"Failed to get message context: {result.get('description')}")
    
    return result


# Example usage and testing sequence
if __name__ == "__main__":
    # Try to get chat ID if not set
    if TEST_CHAT_ID is None:
        TEST_CHAT_ID = get_chat_id(timeout=30)
        if TEST_CHAT_ID is None:
            print("Could not get chat ID. Please either:")
            print("1. Run the script again and send a message to your bot within 30 seconds, or")
            print("2. Set TEST_CHAT_ID manually at the top of the script")
            exit(1)
    
    CHAT_ID = TEST_CHAT_ID
    USER_ID = "TEST_USER_ID"  # For testing get_chat_member
    
    # Test 1: Send a message
    print("\n=== Testing Send Message ===")
    message_response = test_send_message(
        chat_id=CHAT_ID,
        text="Hello! This is a test message."
    )
    
    if "result" in message_response:
        message_id = message_response["result"]["message_id"]
        
        # Test 2: Pin the message we just sent
        print("\n=== Testing Pin Message ===")
        test_pin_message(
            chat_id=CHAT_ID,
            message_id=str(message_id)
        )
        
        # Test 3: Delete the message
        # print("\n=== Testing Delete Message ===")
        # test_delete_message(
        #     chat_id=CHAT_ID,
        #     message_id=str(message_id)
        # )
    
    # Test 4: Create a poll
    print("\n=== Testing Create Poll ===")
    test_create_poll(
        chat_id=CHAT_ID,
        question="How's this test poll?",
        options=["Great!", "Works fine", "Needs improvement"]
    )
    
    # Test 5: Send media
    print("\n=== Testing Send Media ===")
    test_send_media(
        chat_id=CHAT_ID,
        media_url="https://pbs.twimg.com/media/GOQRcbhbIAAofos.jpg",
        media_type="photo",
        caption="Test image"
    )

    # Test 5: Get chat history
    print("\n=== Testing Get Chat History ===")
    test_get_chat_history(chat_id=CHAT_ID)

    # Test 6: Get message context
    if "result" in message_response and "message_id" in message_response["result"]:
        print("\n=== Testing Get Message Context ===")
        test_get_message_context(
            chat_id=CHAT_ID
        )

    # # Test 6: Get chat member info
    # print("\n=== Testing Get Chat Member ===")
    # test_get_chat_member(
    #     chat_id=CHAT_ID,
    #     user_id=USER_ID
    # )




from typing import List, Optional, Callable, Dict
import uuid
from virtuals_sdk.game.worker import Worker
from virtuals_sdk.game.custom_types import Function, FunctionResult, FunctionResultStatus, ActionResponse, ActionType
from virtuals_sdk.game.utils import create_agent, create_workers, post

class Session:
    def __init__(self):
        self.id = str(uuid.uuid4())
        self.function_result: Optional[FunctionResult] = None

    def reset(self):
        self.id = str(uuid.uuid4())
        self.function_result = None


class WorkerConfig:
    def __init__(self,
                 id: str,
                 worker_description: str,
                 get_state_fn: Callable,
                 action_space: List[Function],
                 instruction: Optional[str] = "",
                 ):

        self.id = id  # id or name of the worker
        # worker description for the TASK GENERATOR (to give appropriate tasks) [NOT FOR THE WORKER ITSELF - WORKER WILL STILL USE AGENT DESCRIPTION]
        self.worker_description = worker_description
        self.instruction = instruction
        self.get_state_fn = get_state_fn
        self.action_space = action_space

        # setup get state function with the instructions
        self.get_state_fn = lambda function_result, current_state: {
            "instructions": self.instruction,  # instructions are set up in the state
            # places the rest of the output of the get_state_fn in the state
            **get_state_fn(function_result, current_state),
        }

        self.action_space: Dict[str, Function] = {
            f.get_function_def()["fn_name"]: f for f in action_space
        }


class Agent:
    def __init__(self,
                 api_key: str,
                 name: str,
                 agent_goal: str,
                 agent_description: str,
                 get_agent_state_fn: Callable,
                 workers: Optional[List[WorkerConfig]] = None,
                 ):

        self._base_url: str = "https://game.virtuals.io"
        self._api_key: str = api_key

        # checks
        if not self._api_key:
            raise ValueError("API key not set")

        # initialize session
        self._session = Session()

        self.name = name
        self.agent_goal = agent_goal
        self.agent_description = agent_description

        # set up workers
        if workers is not None:
            self.workers = {w.id: w for w in workers}
        else:
            self.workers = {}
        self.current_worker_id = None

        # get agent/task generator state function
        self.get_agent_state_fn = get_agent_state_fn

        # initialize and set up agent states
        self.agent_state = self.get_agent_state_fn(None, None)

        # create agent
        self.agent_id = create_agent(
            self._base_url, self._api_key, self.name, self.agent_description, self.agent_goal
        )

    def compile(self):
        """ Compile the workers for the agent - i.e. set up task generator"""
        if not self.workers:
            raise ValueError("No workers added to the agent")

        workers_list = list(self.workers.values())

        self._map_id = create_workers(
            self._base_url, self._api_key, workers_list)
        self.current_worker_id = next(iter(self.workers.values())).id

        # initialize and set up worker states
        worker_states = {}
        for worker in workers_list:
            dummy_function_result = FunctionResult(
                action_id="",
                action_status=FunctionResultStatus.DONE,
                feedback_message="",
                info={},
            )
            worker_states[worker.id] = worker.get_state_fn(
                dummy_function_result, self.agent_state)

        self.worker_states = worker_states

        return self._map_id

    def reset(self):
        """ Reset the agent session"""
        self._session.reset()

    def add_worker(self, worker_config: WorkerConfig):
        """Add worker to worker dict for the agent"""
        self.workers[worker_config.id] = worker_config
        return self.workers

    def get_worker_config(self, worker_id: str):
        """Get worker config from worker dict"""
        return self.workers[worker_id]

    def get_worker(self, worker_id: str):
        """Initialize a working interactable standalone worker"""
        worker_config = self.get_worker_config(worker_id)
        return Worker(
            api_key=self._api_key,
            # THIS DESCRIPTION IS THE AGENT DESCRIPTION/CHARACTER CARD - WORKER DESCRIPTION IS ONLY USED FOR THE TASK GENERATOR
            description=self.agent_description,
            instruction=worker_config.instruction,
            get_state_fn=worker_config.get_state_fn,
            action_space=worker_config.action_space,
        )

    def _get_action(
        self,
        function_result: Optional[FunctionResult] = None
    ) -> ActionResponse:

        # dummy function result if None is provided - for get_state_fn to take the same input all the time
        if function_result is None:
            function_result = FunctionResult(
                action_id="",
                action_status=FunctionResultStatus.DONE,
                feedback_message="",
                info={},
            )

        # set up payload
        data = {
            "location": self.current_worker_id,
            "map_id": self._map_id,
            "environment": self.worker_states[self.current_worker_id],
            "functions": [
                f.get_function_def()
                for f in self.workers[self.current_worker_id].action_space.values()
            ],
            "events": {},
            "agent_state": self.agent_state,
            "current_action": (
                function_result.model_dump(
                    exclude={'info'}) if function_result else None
            ),
            "version": "v2",
        }

        # make API call
        response = post(
            base_url=self._base_url,
            api_key=self._api_key,
            endpoint=f"/v2/agents/{self.agent_id}/actions",
            data=data,
        )

        return ActionResponse.model_validate(response)

    def step(self):

        # get next task/action from GAME API
        action_response = self._get_action(self._session.function_result)
        action_type = action_response.action_type

        print("#" * 50)
        print("STEP")
        print(f"Current Task: {action_response.agent_state.current_task}")
        print(f"Action response: {action_response}")
        print(f"Action type: {action_type}")

        # if new task is updated/generated
        if (
            action_response.agent_state.hlp
            and action_response.agent_state.hlp.change_indicator
        ):
            print("New task generated")
            print(f"Task: {action_response.agent_state.current_task}")

        # execute action
        if action_type in [
            ActionType.CALL_FUNCTION,
            ActionType.CONTINUE_FUNCTION,
        ]:
            print(f"Action Selected: {action_response.action_args['fn_name']}")
            print(f"Action Args: {action_response.action_args['args']}")

            if not action_response.action_args:
                raise ValueError("No function information provided by GAME")

            self._session.function_result = (
                self.workers[self.current_worker_id]
                .action_space[action_response.action_args["fn_name"]]
                .execute(**action_response.action_args)
            )

            print(f"Function result: {self._session.function_result}")

            # update worker states
            updated_worker_state = self.workers[self.current_worker_id].get_state_fn(
                self._session.function_result, self.worker_states[self.current_worker_id])
            self.worker_states[self.current_worker_id] = updated_worker_state

        elif action_response.action_type == ActionType.WAIT:
            print("Task ended completed or ended (not possible wiht current actions)")

        elif action_response.action_type == ActionType.GO_TO:
            if not action_response.action_args:
                raise ValueError("No location information provided by GAME")

            next_worker = action_response.action_args["location_id"]
            print(f"Next worker selected: {next_worker}")
            self.current_worker_id = next_worker

        else:
            raise ValueError(
                f"Unknown action type: {action_response.action_type}")

        # update agent state
        self.agent_state = self.get_agent_state_fn(
            self._session.function_result, self.agent_state)

    def run(self):
        self._session = Session()
        while True:
            self.step()



from virtuals_sdk.game import Agent
from virtuals_sdk.functions.telegram import TelegramClient
from virtuals_sdk.functions.farcaster import FarcasterClient
from virtuals_sdk.functions.discord import DiscordClient

# # Create agent with just strings for each component
# agent = Agent(
# 	api_key="985f75776b45337f0a9c2c11ddf7bd795950ee13b4e5074af604762f449114bd",
#     goal="Autonomously analyze crypto markets and provide trading insights",
#     description="HODL-9000: A meme-loving trading bot powered by hopium and ramen",
#     world_info="Virtual crypto trading environment where 1 DOGE = 1 DOGE"
# )

# # list available functions that can be added to agent
# # print(agent.list_available_default_twitter_functions()) 
# # agent.list_available_default_twitter_functions()
# # agent.use_default_twitter_functions(["wait", "reply_tweet"])

# print("here")
# # response = agent.simulate_twitter(session_id="session-twitter")
# # print(response)

# ### TEST TELEGRAM FUNCTIONS
# # get telegram custom functions through client
# tg_client = TelegramClient(bot_token="7210003118:AAGDgu4PgMJEwY892B_U0wyfZ0xCm-sMOGk")

# # get available telegram functions
# print(tg_client.available_functions)

# # get telegram custom functions through client
# reply_message_fn = tg_client.get_function("send_message")
# create_poll_fn = tg_client.get_function("create_poll")
# pin_message_fn = tg_client.get_function("pin_message")
# # set_chat_title_fn = tg_client.get_function("set_chat_title")

# reply_message_fn("5394289251", "Hello World")
# create_poll_fn("5394289251", "What is your favorite color?", ["Red", "Blue", "Green"], "True")
# pin_message_fn("5394289251", "82", "True")


# ### TEST FARCASTER FUNCTIONS
# NEYNAR_API_KEY = ""
# NEYNAR_SIGNER_UUID = ""
# farcaster_client = FarcasterClient(api_key=NEYNAR_API_KEY,
#                                    signer_uuid=NEYNAR_SIGNER_UUID)

# # get available farcaster functions
# print(farcaster_client.available_functions)

# # get farcaster custom functions through client
# post_cast_fn = farcaster_client.get_function("post_cast")
# reply_to_cast_fn = farcaster_client.get_function("reply_to_cast")
# get_trending_casts_fn = farcaster_client.get_function("get_trending_casts")
# search_casts_fn = farcaster_client.get_function("search_casts")



### TEST DISCORD FUNCTIONS
discord_client = DiscordClient(bot_token="MTMxODQ1MDE2OTYxNTc0OTE4MA.Gr4v3O.JC_lXLQ_EvVMd-C_BeC0xwrFLZDQE5DZqtndXk")
dis_send_message_fn = discord_client.get_function("send_message")
dis_reaction_fn = discord_client.get_function("add_reaction")
dis_pin_message_fn = discord_client.get_function("pin_message")
dis_delete_message_fn = discord_client.get_function("delete_message")

channel_id = "1318449474195947522"
dis_send_message_fn(channel_id, "Hello World")

message_id = "1321715872863092736"
dis_reaction_fn(channel_id, message_id, "😊")

dis_pin_message_fn(channel_id, message_id)

message_id = "1321732428250877984"
dis_delete_message_fn(channel_id, message_id)